import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.ticker as ticker
import numpy as np
from .inputs import parse_coil

# Plotting routines


def plot_fields(fields: np.ndarray, positions: np.ndarray,
                which_plane='z', level=0, num_contours=50) -> None:
    # Ryan's Magic Code
    '''
    Plots the set of Bfields in the given region, at the specified resolutions.

    Bfields: A 4D array of the Bfield.
    box_size: (x, y, z) dimensions of the box in cm
    start_point: (x, y, z) = (0, 0, 0) = bottom left corner position of the box AKA the offset
    vol_resolution: Division of volumetric meshgrid (generate a point every volume_resolution cm)
    which_plane: Plane to plot on, can be "x", "y" or "z"
    level : The "height" of the plane. For instance the Z = 5 plane would have a level of 5
    num_contours: THe amount of contours on the contour plot.

    '''
    # filled contour plot of Bx, By, and Bz on a chosen slice plane
    X = positions[:, 0, 0, 0]
    Y = positions[0, :, 0, 1]
    Z = positions[0, 0, :, 2]

    if which_plane == 'x':

        converted_level = np.where(X >= level)

        B_sliced = [fields[converted_level[0]
                            [0], :, :, i].T for i in range(3)]
        x_label, y_label = "y", "z"
        x_array, y_array = Y, Z
    elif which_plane == 'y':
        converted_level = np.where(Y >= level)
        B_sliced = [fields[:, converted_level[0]
                            [0], :, i].T for i in range(3)]
        x_label, y_label = "x", "z"
        x_array, y_array = X, Z
    else:
        converted_level = np.where(Z >= level)
        B_sliced = [
            fields[:, :, converted_level[0][0], i].T for i in range(3)]
        x_label, y_label = "x", "y"
        x_array, y_array = X, Y

    Bmin, Bmax = np.amin(B_sliced), np.amax(B_sliced)

    component_labels = ['x', 'y', 'z']
    fig, axes = plt.subplots(nrows=1, ncols=4, figsize=(10, 5))
    axes[0].set_ylabel(y_label + " (cm)")

    for i in range(3):
        contours = axes[i].contourf(x_array, y_array, B_sliced[i],
                                    vmin=Bmin, vmax=Bmax,
                                    cmap=cm.magma, levels=num_contours)
        axes[i].set_xlabel(x_label + " (cm)")
        axes[i].set_title(
            "$\\mathcal{B}$" +
            "$_{}$".format(
                component_labels[i]))

    axes[3].set_aspect(20)
    fig.colorbar(contours, cax=axes[3], extend='both')

    plt.tight_layout()

    plt.show()


def plot_coil(*input_filenames):
    '''
    Plots one or more coils in space.

    input_filenames: Name of the files containing the coils.
    Should be formatted appropriately.
    '''
    fig = plt.figure()
    tick_spacing = 2
    ax = fig.add_subplot(111, projection='3d')
    ax.set_xlabel("$x$ (cm)")
    ax.set_ylabel("$y$ (cm)")
    ax.set_zlabel("$z$ (cm)")

    for input_filename in input_filenames:
        coil_points = np.array(parse_coil(input_filename))

        ax.plot3D(coil_points[:, 0], coil_points[:, 1],
                  coil_points[:, 2], lw=2)
    for axis in [ax.xaxis, ax.yaxis, ax.zaxis]:
        axis.set_major_locator(ticker.MultipleLocator(tick_spacing))
    plt.tight_layout()
    plt.show()


def plot_quiver(fields: np.ndarray, positions: np.ndarray) -> None:
    '''
    3D Quiver plot of the thing in a volume
    '''

    ax = plt.figure().add_subplot(projection='3d')

    mags = np.linalg.norm(fields, axis=3)

    q = ax.quiver(positions[:,:,:,0], positions[:,:,:,1], positions[:,:,:,2],
                  fields[:,:,:,0], fields[:,:,:,1], fields[:,:,:,2], cmap="coolwarm", normalize=True)
    q.set_array(mags.ravel())  # Ravel the array
    plt.colorbar(q, cmap="coolwarm")

    plt.show()



'''
Biot-Savart Magnetic Field Calculator v5.0
Mingde Yin
Ryan Zazo

June 2021

All lengths are in cm, B-field is in G
'''

import numpy as np
from .inputs import parse_coil

'''
Feature Wishlist:
    improve plot_coil with different colors for different values of current

    DONE
    accelerate integrator to use meshgrids directly instead of a layer of for loop

    DONE
    get parse_coil to use vectorized function instead of for loop
'''


def slice_coil(coil: np.ndarray, currents: np.ndarray,
               steplength: float) -> "tuple[np.ndarray, np.ndarray]":
    '''
    Slices a coil into pieces of size steplength,
    synchronizing current bits as well.
    If the coil is already sliced into pieces
    smaller than that, this does nothing.
    '''

    def interpolate_points(p1, p2, current, parts):
        '''
        Produces a series of linearly spaced points between two given points,
        with constant current

        e.g. [0, 2, 1], [3, 4, 2], 1, parts=2
        >>> [[0, 2, 1], [1.5, 3, 1.5], [3, 4, 2]], [1, 1, 1]
        '''
        coil_points = np.linspace(p1, p2, parts + 1)
        currents = current * np.ones(parts + 1)
        # TODO see if this actually works properly

        return coil_points, currents

    # fill column with dummy data, we will remove this later.
    newcoil = np.zeros((1, 3))
    newcurrents = np.zeros(1)

    segment_starts = coil[:-1, :]
    segment_ends = coil[1:, :]
    # determine start and end of each segment

    segments = segment_ends - segment_starts
    segment_lengths = np.linalg.norm(segments, axis=1)
    # create segments; determine start and end of each segment,
    # as well as segment lengths

    # chop up into smaller bits (elements)

    stepnumbers = (segment_lengths / steplength).astype(int)
    # determine how many steps we must chop each segment into

    for i in range(segments.shape[0]):
        newrows, newcurrs = interpolate_points(
            segment_starts[i], segment_ends[i], currents[i], stepnumbers[i])
        # set of new interpolated points to feed in
        newcoil = np.vstack((newcoil, newrows))
        newcurrents = np.hstack((newcurrents, newcurrs))

    return newcoil[1:], newcurrents[1:]  # return non-dummy columns


def calculate_field(coil: np.ndarray, current: np.ndarray,
                    positions: np.ndarray) -> np.ndarray:
    '''
    Calculates free space magnetic flux density as a result
    of some position and current x, y, z, I
    [In the same coordinate system as the coil]

    Coil with N segments defined by N+1 points
    Space with U x V x W dimensions

    Coil: Input Coil Positions, already
    sub-divided into small pieces using slice_coil
    Array shape (N, 3)

    Positions: (U x V x W x 3) array of positions in cm

    Output B-field is a 3-D vector in units of G
    '''
    FACTOR = 0.1
    # mu_0 / 4pi when lengths are in cm, and B-field is in G, current in A

    midpoints = (coil[1:] + coil[:-1])/2  # midpoints of each coil position
    dl = np.diff(coil, axis=0)  # dl row vectors for each segment
    R_Rprime = positions[:, :, :, np.newaxis, :] - \
        midpoints[np.newaxis, np.newaxis, np.newaxis, :]
   
    # R - R' term, of shape(U, V, W, N, 3)
    mags = np.linalg.norm(R_Rprime, axis=-1)

    elemental_integrands = FACTOR * \
        current[np.newaxis, np.newaxis, np.newaxis, :-1, np.newaxis] * \
        np.cross(dl, R_Rprime) / mags[:, :, :, :, np.newaxis]**3
    # Evaluate the integrand (U, V, W, N, 3) using BSL
    # BSL is current * mu/4pi * dl x (R-R') / |R-R'|^3
    # The "area" underneath each rectangle

    return np.sum(elemental_integrands, axis=-2)  # sum of all "areas"


def generate_positions(box_size: tuple,
                       start_point: tuple,
                       vol_resolution: float) -> np.ndarray:
    '''
    Generates positions at which magnetic field is evaluated
    Box Size: 3 element tuple of form (u, v, w),
    where u, v, w are the x, y, z dimensions of the box in cm

    Start Point: 3 element tuple of form (a, b, c)
    where a, b, c are the x, y, z coordinates
    of the starting corner of the box in cm

    Vol Resolution: Resolution of the grid in cm.
    '''
    x = np.linspace(start_point[0],
                    box_size[0] + start_point[0],
                    int(box_size[0] / vol_resolution) + 1)
    y = np.linspace(start_point[1],
                    box_size[1] + start_point[1],
                    int(box_size[1] / vol_resolution) + 1)
    z = np.linspace(start_point[2],
                    box_size[2] + start_point[2],
                    int(box_size[2] / vol_resolution) + 1)
    # Generate points at regular spacing, incl. end points
    xx, yy, zz, = np.meshgrid(x, y, z, indexing="ij")

    return np.transpose(np.array([xx, yy, zz]), axes=(1, 2, 3, 0))
    # Reorder axes so that the last axis is the (x, y, z) row vector


def produce_target_volume(coil: np.ndarray, current: np.ndarray,
                          box_size: tuple, start_point: tuple,
                          vol_resolution: float) -> "tuple[np.ndarray, np.ndarray]":
    '''
    Generates a set of field vector values for each tuple (x, y, z) in the box.
​
    Coil, Current: Input Coil Positions and Currents in format specified above,
    already sub-divided into small pieces

    box_size: (x, y, z) dimensions of the box in cm
    start_point: (x, y, z) = (0, 0, 0) = bottom left corner position of the box
    vol_resolution: Spatial resolution (in cm)

    Returns: B fields and Positions of Evaluation
    '''
    positions = generate_positions(box_size, start_point, vol_resolution)
    return calculate_field(coil, current, positions), positions


def get_field_vector(targetVolume, position, start_point, volume_resolution):
    # REDO
    '''
    Returns the B vector [Bx, By, Bz] components in a generated Target Volume at a given position tuple (x, y, z) in a coordinate system

    start_point: (x, y, z) = (0, 0, 0) = bottom left corner position of the box
    volume_resolution: Division of volumetric meshgrid (generate a point every volume_resolution cm)
    '''
    relativePosition = (
        (np.array(position) -
         np.array(start_point)) /
        volume_resolution).astype(int)
    # adjust to the meshgrid's system

    if (relativePosition < 0).any():
        return ("ERROR: Out of bounds! (negative indices)")

    try:
        return targetVolume[relativePosition[0],
                            relativePosition[1], relativePosition[2], :]
    except BaseException:
        return ("ERROR: Out of bounds!")
    # basic error checking to see if you actually got a correct input/output


def write_target_volume(input_filename: str, output_filename: str,
                        box_size: tuple, start_point: tuple,
                        coil_resolution: float = 1,
                        volume_resolution: float = 1) -> None:
    '''
    Takes a coil specified in input_filename, generates a target volume, and saves the generated target volume to output_filename.

    box_size: (x, y, z) dimensions of the box in cm
    start_point: (x, y, z) = (0, 0, 0) = bottom left corner position of the box AKA the offset
    coil_resolution: How long each coil subsegment should be
    volume_resolution: Division of volumetric meshgrid (generate a point every volume_resolution cm)
    '''
    coil, current = slice_coil(*parse_coil(input_filename), coil_resolution)

    fields, positions = produce_target_volume(
        coil, current, box_size, start_point, volume_resolution)

    with open(f"{output_filename}_fields", "wb") as f:
        np.save(f, fields)

    with open(f"{output_filename}_positions", "wb") as f:
        np.save(f, positions)
    # stored in standard numpy pickle form
